"""
SPARTN Bitmask, Lookup and Decode Constants

Created on 10 Feb 2023

Information Sourced from https://www.spartnformat.org/download/
(available in the public domain) © 2021 u-blox AG. All rights reserved.

:author: semuadmin
"""

SATBITMASKKEY = {
    "GPS": "SF011",
    "GLO": "SF012",
    "GAL": "SF093",
    "BEI": "SF094",
    "QZS": "SF095",
}
"""Satellite PRN bitmask keys"""

SATIODEKEY = {
    "GPS": "SF018",
    "GLO": "SF019",
    "GAL": "SF099",
    "BEI": "SF100",
    "QZS": "SF101",
}
"""Satellite IODE keys"""

SF011_ENUM = {
    0: 32,
    1: 44,
    2: 56,
    3: 64,
}
"""GPS satellite mask length (leftmost 2 bits of SF011)"""

SF012_ENUM = {
    0: 24,
    1: 36,
    2: 48,
    3: 63,
}
"""GLONASS satellite mask length (leftmost 2 bits of SF012)"""

SF093_ENUM = {
    0: 36,
    1: 45,
    2: 54,
    3: 64,
}
"""Galileo satellite mask length (leftmost 2 bits of SF093)"""

SF094_ENUM = {
    0: 37,
    1: 46,
    2: 55,
    3: 64,
}
"""BDS satellite mask length (leftmost 2 bits of SF094)"""

SF095_ENUM = {
    0: 10,
    1: 40,
    2: 48,
    3: 64,
}
"""QZSS satellite mask length (leftmost 2 bits of SF095)"""

SATBITMASKLEN = {
    SATBITMASKKEY["GPS"]: list(SF011_ENUM.values()),
    SATBITMASKKEY["GLO"]: list(SF012_ENUM.values()),
    SATBITMASKKEY["GAL"]: list(SF093_ENUM.values()),
    SATBITMASKKEY["BEI"]: list(SF094_ENUM.values()),
    SATBITMASKKEY["QZS"]: list(SF095_ENUM.values()),
}
"""Satellite PRN bitmask lengths (PRN values are bitmask position + 1)"""

PBBITMASKKEY = {
    "GPS": "SF025",
    "GLO": "SF026",
    "GAL": "SF102",
    "BEI": "SF103",
    "QZS": "SF104",
}
"""Phase bias bitmask keys"""

PBBITMASKLEN = {
    PBBITMASKKEY["GPS"]: (
        [6, 11],
        {
            0: "L1C",
            1: "L2W",
            2: "L2L",
            3: "L5Q",
            # 4-10: spare phase bias
        },
    ),
    PBBITMASKKEY["GLO"]: (
        [5, 9],
        {
            0: "L1C",
            1: "L2C",
            # 2-8: spare phase bias
        },
    ),
    PBBITMASKKEY["GAL"]: (
        [8, 15],
        {
            0: "L1C",
            1: "L5Q",
            2: "L7Q",
            # 3-14: spare phase bias
        },
    ),
    PBBITMASKKEY["BEI"]: (
        [8, 15],
        {
            0: "L2I",
            1: "L5P",
            2: "L7I",
            3: "L6I",
            4: "L1P",
            5: "L7P",
            6: "L8P",
            # 7-14: spare phase bias
        },
    ),
    PBBITMASKKEY["QZS"]: (
        [6, 11],
        {
            0: "L1C",
            1: "L2L",
            2: "L5Q",
            # 3-10: spare phase bias
        },
    ),
}
"""Phase bias bitmask lengths and enumerations"""

CBBITMASKKEY = {
    "GPS": "SF027",
    "GLO": "SF028",
    "GAL": "SF105",
    "BEI": "SF106",
    "QZS": "SF107",
}
"""Code bias bitmask keys"""

CBBITMASKLEN = {
    CBBITMASKKEY["GPS"]: (
        [6, 11],
        {
            0: "C1C",
            1: "C2W",
            2: "C2L",
            3: "C5Q",
            # 4-10: spare code bias
        },
    ),
    CBBITMASKKEY["GLO"]: (
        [5, 9],
        {
            0: "C1C",
            1: "C2C",
            # 2 to 8 : spare code bias
        },
    ),
    CBBITMASKKEY["GAL"]: (
        [8, 15],
        {
            0: "C1C",
            1: "C5Q",
            2: "C7Q",
            # 3-14: spare code bias
        },
    ),
    CBBITMASKKEY["BEI"]: (
        [8, 15],
        {
            0: "C2I",
            1: "C5P",
            2: "C7I",
            3: "C6I",
            4: "C1P",
            5: "C7P",
            6: "C8P",
            # 7-14: spare code bias
        },
    ),
    CBBITMASKKEY["QZS"]: (
        [6, 11],
        {
            0: "C1C",
            1: "C2L",
            2: "C5Q",
            # 3-10: spare code bias
        },
    ),
}
"""Code bias bitmask lengths and enumerations"""

ALN_ENUM = {0: 8, 1: 12, 2: 16, 3: 32, 4: 64}
"""Embedded authorisation length enumeration"""

SF015_ENUM = SF022_ENUM = {
    0: "0 secs",
    1: "1 secs",
    2: "5 secs",
    3: "10 secs",
    4: "30 secs",
    5: "60 secs",
    6: "120 secs",
    7: "320 secs",
}
"""Continuity indicator enumeration"""

SF024_ENUM = {
    0: "unknown",
    1: "0.01 m",
    2: "0.02 m",
    3: "0.05 m",
    4: "0.1 m",
    5: "0.3 m",
    6: "1.0 m",
    7: "> 1.0 m",
}
"""User range error (URE) enumeration"""

SF042_ENUM = {
    0: "unknown",
    1: "<= 0.010 m",
    2: "<= 0.020 m",
    3: "<= 0.040 m",
    4: "<= 0.080 m",
    5: "<= 0.160 m",
    6: "<= 0.320 m",
    7: "> 0.320 m",
}
"""Troposphere quality enumeration"""

SF044_ENUM = {
    0: "Troposphere small coefficient block",
    1: "Troposphere large coefficient block",
}
"""Troposphere polynomial coefficient size indicator"""

SF051_ENUM = {
    0: "Troposphere small residual",
    1: "Tropospherelarge residual",
}
"""Troposphere residual field size"""

SF055_ENUM = {
    0: "Unknown",
    1: "<= 0.03 TECU",
    2: "<= 0.05 TECU",
    3: "<= 0.07 TECU",
    4: "<= 0.14 TECU",
    5: "<= 0.28 TECU",
    6: "<= 0.56 TECU",
    7: "<= 1.12 TECU",
    8: "<= 2.24 TECU",
    9: "<= 4.48 TECU",
    10: "<= 8.96 TECU",
    11: "<= 17.92 TECU",
    12: "<= 35.84 TECU",
    13: "<= 71.68 TECU",
    14: "<= 143.36 TECU",
    15: "> 143.36 TECU",
}
"""Ionosphere quality enumeration"""

SF056_ENUM = {
    0: "Ionosphere small coefficient block",
    1: "Ionosphere large coefficient block",
}
"""Ionosphere polynomial coefficient size indicator"""

SF063_ENUM = {
    0: "Ionosphere small residual",
    1: "Ionosphere medium residual",
    2: "Ionosphere large residual",
    3: "Ionosphere extra large residual",
}
"""Ionosphere residual field size enumeration"""

SF070_ENUM = {
    0: "350 km",
    1: "400 km",
    2: "450 km",
    3: "500 km",
}
"""Ionosphere shell height enumeration"""

SF077_ENUM = SF078_ENUM = {
    0: "2.5 deg",
    1: "5.0 deg",
    2: "10.0 deg",
    3: "15.0 deg",
}
"""BPAC area latitude/longitude grid node spacing enumeration"""

SF081_ENUM = {
    0: "small VTEC residual",
    1: "large VTEC residual",
}
"""VTEC size indicator"""

SF085_ENUM = {
    0: "AES",
    1: "ChaCha12",
    2: "ChaCha20",
    # 3-15 : TBD
}
"""Encryption Type enumeration"""

SF087_ENUM = {  # no bits
    0: 96,
    1: 128,
    2: 192,
    3: 256,
    4: 512,
    # 5-15 : TBD
}
"""Key length enumeration"""

SF090_ENUM = {
    0: "none",
    1: "Ed25519",
    2: "SHA-2",
    3: "SHA-3",
    # 4-15 : TBD
}
"""Group Authentication Type enumeration"""

SF091_ENUM = {  # no bits
    0: 32,
    1: 64,
    2: 96,
    3: 128,
    4: 192,
    5: 256,
    6: 512,
    # 7-15 : TBD
}
"""Computed Authentication Data (CAD) Length enumeration"""

SF096_ENUM = {
    0: "Galileo F/NAV",
    1: "Galileo I/NAV",
    2: "Galileo C/NAV",
    # 3-7: TBD
}
"""Galileo ephemeris type"""

SF097_ENUM = {
    0: "D1 Nav (B1I)",
    1: "D2 Nav (B1I)",
    2: "D1 Nav (B3I)",
    3: "D2 Nav (B3I)",
    4: "B-CNAV1",
    5: "B-CNAV2",
    # 6-15: TBD
}
"""BDS ephemeris type"""

SF098_ENUM = {
    0: "LNAV (L1C/A)",
    1: "CNAV2 (L1C)",
    2: "CNAV (L2C,L5)",
    # 3-7: TBD
}
"""QZSS ephemeris type"""



"""
SPARTNReader class.

The SPARTNReader class will parse individual SPARTN messages
from any binary stream containing *solely* SPARTN data e.g. an
MQTT `/pp/ip` topic.

Information sourced from https://www.spartnformat.org/download/
(available in the public domain)
© 2021 u-blox AG. All rights reserved.

SPARTN 1X transport layer bit format:

+-----------+------------+-------------+-------------+------------+-----------+
| preamble  | framestart | payload     |  payload    | embedded   |   crc     |
|           |            | descriptor  |             | auth data  |           |
+===========+============+=============+=============+============+===========+
| 8 bits    |  24 bits   | 32-64 bits  | 8-8192 bits | 0-512 bits | 8-32 bits |
| 0x73 's'  |            |             |             |            |           |
+-----------+------------+-------------+-------------+------------+-----------+

NB Use of gnssTimeTag for message decryption:

The SPARTN protocol requires a key and basedate to calculate the Initialisation
Vector (IV) for encrypted messages (eaf=1). The key is provided by the 
SPARTN service provider. The basedate is derived in one of two ways:

1. For messages with unambiguous 32-bit gnssTimeTag values (timeTagtype = 1),
   the basedate is the gnssTimeTag. No other information is needed.

2. For messages with ambiguous 16-bit gnssTimeTag values (timeTagtype = 0),
   the basedate can be derived from a 32-bit gnssTimeTag for the same message
   subtype (GPS, GLO, etc.) from the same datastream, or provided as an external
   parameter. SPARTNReader will accumulate any 32-bit gnssTimeTag in the incoming
   datastream for use in decryption.


Created on 10 Feb 2023

:author: semuadmin
:copyright: SEMU Consulting © 2023
:license: BSD 3-Clause
"""

# pylint: disable=invalid-name too-many-instance-attributes

from logging import getLogger
from os import getenv
from socket import socket

from pyspartn.exceptions import (
    ParameterError,
    SPARTNDecryptionError,
    SPARTNMessageError,
    SPARTNParseError,
    SPARTNStreamError,
    SPARTNTypeError,
)
from pyspartn.socket_wrapper import SocketWrapper
from pyspartn.spartnhelpers import bitsval, valid_crc
from pyspartn.spartnmessage import SPARTNMessage
from pyspartn.spartntables import ALN_ENUM
from pyspartn.spartntypes_core import ERRLOG, ERRRAISE, SPARTN_PREB, VALCRC


class SPARTNReader:
    """
    SPARTNReader class.
    """

    def __init__(
        self,
        datastream,
        validate: int = VALCRC,
        quitonerror: int = ERRLOG,
        decode: bool = False,
        key: str = None,
        basedate: object = None,
        bufsize: int = 4096,
        errorhandler: object = None,
        timetags: dict = None,
    ):
        """Constructor.

        :param datastream stream: input data stream
        :param int validate: VALCRC (1) = validate CRC, VALNONE (1) = ignore invalid CRC (1)
        :param int quitonerror: ERROR_IGNORE (0) = ignore,  ERROR_LOG (1) = log and continue,
            ERROR_RAISE (2) = (re)raise (1)
        :param bool decode: decrypt and decode payload (False)
        :param str key: decryption key as hexadecimal string (None)
        :param object basedate: decryption basedate as datetime or 32-bit gnssTimeTag as
           integer (None). If basedate = TIMEBASE, SPARTNMessage will use timetags argument
        :param int bufsize: socket recv buffer size (4096)
        :param int errorhandler: error handling object or function (None)
        :param dict timetags: dict of decryption timetags in format {0: 442626332, 1: 449347321,
            2: 412947745} where key = msgSubtype (0=GPS, 1=GLO, etc) and value = gnssTimeTag (None)
        :raises: ParameterError if invalid parameters
        :raises: SPARTNDecryptionError if unable to decrypt message
            using key and basedate/timetags provided
        :raises: SPARTN***Error if unable to parse message
        """
        # pylint: disable=too-many-arguments

        self._logger = getLogger(__name__)
        if isinstance(datastream, socket):
            self._stream = SocketWrapper(datastream, bufsize=bufsize)
        else:
            self._stream = datastream
        self.key = getenv("MQTTKEY", None) if key is None else key
        self._validate = validate
        self._quitonerror = quitonerror
        self._errorhandler = errorhandler
        self._decode = decode
        self._key = key
        self._basedate = basedate
        # accumlated array of 32-bit gnssTimeTag from datastream
        self._timetags = {} if timetags is None else timetags

        if self._decode and self._key is None:
            raise ParameterError("Key must be provided if decoding is enabled")

    def __iter__(self):
        """Iterator."""

        return self

    def __next__(self) -> tuple:
        """
        Return next item in iteration.

        :return: tuple of (raw_data as bytes, parsed_data as SPARTNessage or None if error)
        :rtype: tuple
        :raises: StopIteration
        """

        (raw_data, parsed_data) = self.read()
        if raw_data is None and parsed_data is None:
            raise StopIteration
        return (raw_data, parsed_data)

    def read(self) -> tuple:
        """
        Read a single SPARTN message from the stream buffer
        and return both raw and parsed data.

        The 'quitonerror' flag determines whether to raise, log or ignore
        parsing errors. If error and quitonerror = 1, the 'parsed' value
        will contain the error message.

        :return: tuple of (raw_data as bytes, parsed_data as SPARTNMessage)
        :rtype: tuple
        :raises: SPARTN***Error if error during parsing
        """

        parsing = True
        raw_data = parsed_data = None

        while parsing:  # loop until end of valid message or EOF
            try:
                raw_data = None
                parsed_data = None
                byte1 = self._read_bytes(1)  # read the first byte
                # if not SPARTN, discard and continue
                if byte1 != SPARTN_PREB:
                    raise SPARTNParseError(f"Unknown protocol {byte1}")
                (raw_data, parsed_data) = self._parse_spartn(byte1)
                parsing = False

            except EOFError:
                return (None, None)
            except (
                SPARTNParseError,
                SPARTNMessageError,
                SPARTNTypeError,
                SPARTNStreamError,
                SPARTNDecryptionError,
            ) as err:
                if self._quitonerror:
                    self._do_error(err)
                continue

        return (raw_data, parsed_data)

    def _parse_spartn(self, preamble: bytes) -> tuple:
        """
        Parse any SPARTN data in the stream. The structure of the transport layer
        depends on encryption type, GNSS timetag format and CRC format.

        :param preamble hdr: preamble of SPARTN message
        :return: tuple of (raw_data as bytes, parsed_stub as SPARTNMessage)
        :rtype: tuple
        :raises: SPARTN...Error if CRC invalid or other parsing error
        """
        # pylint: disable=unused-variable

        framestart = self._read_bytes(3)
        # msgType = bitsval(framestart, 0, 7)
        nData = bitsval(framestart, 7, 10)
        eaf = bitsval(framestart, 17, 1)
        crcType = bitsval(framestart, 18, 2)
        # frameCrc = bitsval(framestart, 20, 4)

        payDesc = self._read_bytes(4)
        # msgSubtype denotes constellation - GPS, GLO, GAL, etc.
        msgSubtype = bitsval(payDesc, 0, 4)
        timeTagtype = bitsval(payDesc, 4, 1)
        if timeTagtype:
            payDesc += self._read_bytes(2)
        gtlen = 32 if timeTagtype else 16
        gnssTimeTag = bitsval(payDesc, 5, gtlen)
        # store 32-bit timetag for this subtype for later use in decryption
        if timeTagtype == 1:
            self._timetags[msgSubtype] = gnssTimeTag
        # solutionId = bitsval(payDesc, gtlen + 5, 7)
        # solutionProcId = bitsval(payDesc, gtlen + 12, 4)
        authInd = 0
        embAuthLen = 0
        if eaf:
            payDesc += self._read_bytes(2)
            # encryptionId = bitsval(payDesc, gtlen + 16, 4)
            # encryptionSeq = bitsval(payDesc, gtlen + 20, 6)
            authInd = bitsval(payDesc, gtlen + 26, 3)
            embAuthLen = bitsval(payDesc, gtlen + 29, 3)
        payload = self._read_bytes(nData)
        embAuth = b""
        if authInd > 1:
            aln = ALN_ENUM.get(embAuthLen, 0)
            embAuth = self._read_bytes(aln)
        crcb = self._read_bytes(crcType + 1)
        crc = int.from_bytes(crcb, "big")

        # validate CRC
        core = framestart + payDesc + payload + embAuth
        raw_data = preamble + core + crcb
        if self._validate & VALCRC:
            if not valid_crc(core, crc, crcType):
                raise SPARTNParseError(f"Invalid CRC {crc}")

        parsed_data = self.parse(
            raw_data,
            validate=self._validate,
            decode=self._decode,
            key=self._key,
            basedate=self._basedate,
            timetags=self.timetags,
        )
        return (raw_data, parsed_data)

    def _read_bytes(self, size: int) -> bytes:
        """
        Read a specified number of bytes from stream.

        :param int size: number of bytes to read
        :return: bytes
        :rtype: bytes
        :raises: EOFError if stream ends prematurely
        """

        data = self._stream.read(size)
        if len(data) == 0:  # EOF
            raise EOFError()
        if 0 < len(data) < size:  # truncated stream
            raise SPARTNStreamError(
                "Serial stream terminated unexpectedly. "
                f"{size} bytes requested, {len(data)} bytes returned."
            )
        return data

    def _do_error(self, err: Exception):
        """
        Handle error.

        :param Exception err: error message
        :raises: Exception if quitonerror = 2
        """

        if self._quitonerror == ERRRAISE:
            raise err from err
        if self._quitonerror == ERRLOG:
            # pass to error handler if there is one
            if self._errorhandler is None:
                self._logger.error(err)
            else:
                self._errorhandler(err)

    @property
    def datastream(self) -> object:
        """
        Getter for stream.

        :return: data stream
        :rtype: object
        """

        return self._stream

    @property
    def timetags(self) -> dict:
        """
        Getter for accumulated 32-bit gnssTimeTag time tags from data stream.

        Can be used as a source of decryption basedate for each
        msgSubtype (i.e. GNSS constellation) if no other basedate
        is supplied.

        :return: dict of gnssTimeTag from data stream (key is msgSubtype)
        :rtype: dict
        """

        return self._timetags

    @staticmethod
    def parse(
        message: bytes,
        validate: int = VALCRC,
        decode: bool = False,
        key: str = None,
        basedate: object = None,
        timetags: dict = None,
    ) -> SPARTNMessage:
        """
        Parse SPARTN message to SPARTNMessage object.

        :param bytes message: SPARTN raw message bytes
        :param int validate: 0 = ignore invalid CRC, 1 = validate CRC (1)
        :param int decode: decode payload True/False
        :param str key: decryption key (required if decode = 1)
        :param object basedate: basedate as datetime or 32-bit gnssTimeTag as integer (None)
        :param dict timetags: dict of accumulated gnssTimeTags from data stream (None)
        :return: SPARTNMessage object
        :rtype: SPARTNMessage
        :raises: SPARTN...Error (if data stream contains invalid data or unknown message type)
        """
        # pylint: disable=unused-argument

        if decode and key is None:
            raise ParameterError("Key must be provided if decoding is enabled")

        return SPARTNMessage(
            transport=message,
            validate=validate,
            decode=decode,
            key=key,
            basedate=basedate,
            timetags=timetags,
        )
